import os
import subprocess

if __name__ == "__main__":
    # Building the frontend is optional
    if os.getenv('SKIP_BUILD_FRONTEND') is None and os.path.exists('package.json'):
        # When building the backend, make sure the frontend is built first
        subprocess.run(['yarn', 'install'], check=True)
        subprocess.run(['yarn', 'build', '--outDir', 'yacv_server/frontend'], check=True)



import hashlib
import io
import re
from typing import List, Dict, Tuple, Union

from OCP.BRep import BRep_Tool
from OCP.BRepAdaptor import BRepAdaptor_Curve
from OCP.GCPnts import GCPnts_TangentialDeflection
from OCP.TopExp import TopExp
from OCP.TopLoc import TopLoc_Location
from OCP.TopTools import TopTools_IndexedMapOfShape
from OCP.TopoDS import TopoDS_Face, TopoDS_Edge, TopoDS_Shape, TopoDS_Vertex
from build123d import Shape, Vertex, Face, Location
from pygltflib import GLTF2

from yacv_server.cad import CADCoreLike
from yacv_server.gltf import GLTFMgr
from yacv_server.mylogger import logger


def tessellate(
        cad_like: CADCoreLike,
        tolerance: float = 0.1,
        angular_tolerance: float = 0.1,
        faces: bool = True,
        edges: bool = True,
        vertices: bool = True,
) -> GLTF2:
    """Tessellate a whole shape into a list of triangle vertices and a list of triangle indices."""
    mgr = GLTFMgr()

    if isinstance(cad_like, TopLoc_Location):
        mgr.add_location(Location(cad_like))

    elif isinstance(cad_like, TopoDS_Shape):
        shape = Shape(cad_like)

        # Perform tessellation tasks
        edge_to_faces: Dict[TopoDS_Edge, List[TopoDS_Face]] = {}
        vertex_to_faces: Dict[TopoDS_Vertex, List[TopoDS_Face]] = {}
        if faces:
            for face in shape.faces():
                _tessellate_face(mgr, face.wrapped, tolerance, angular_tolerance)
                if edges:
                    for edge in face.edges():
                        edge_to_faces[edge.wrapped] = edge_to_faces.get(edge.wrapped, []) + [face.wrapped]
                if vertices:
                    for vertex in face.vertices():
                        vertex_to_faces[vertex.wrapped] = vertex_to_faces.get(vertex.wrapped, []) + [face.wrapped]
        if edges:
            for edge in shape.edges():
                _tessellate_edge(mgr, edge.wrapped, edge_to_faces.get(edge.wrapped, []), angular_tolerance,
                                 angular_tolerance)
        if vertices:
            for vertex in shape.vertices():
                _tessellate_vertex(mgr, vertex.wrapped, vertex_to_faces.get(vertex.wrapped, []))

    return mgr.gltf


def _tessellate_face(
        mgr: GLTFMgr,
        ocp_face: TopoDS_Face,
        tolerance: float = 1e-3,
        angular_tolerance: float = 0.1
):
    face = Shape(ocp_face)
    face.mesh(tolerance, angular_tolerance)
    poly = BRep_Tool.Triangulation_s(face.wrapped, TopLoc_Location())
    if poly is None:
        logger.warn("No triangulation found for face")
        return GLTF2()
    tri_mesh = face.tessellate(tolerance, angular_tolerance)

    # Get UV of each face from the parameters
    uv = [
        (v.X(), v.Y())
        for v in (poly.UVNode(i) for i in range(1, poly.NbNodes() + 1))
    ]

    vertices = [(v.X, v.Y, v.Z) for v in tri_mesh[0]]
    indices = tri_mesh[1]
    mgr.add_face(vertices, indices, uv)


def _push_point(v: Tuple[float, float, float], faces: List[TopoDS_Face]) -> Tuple[float, float, float]:
    # Use the connected faces to push edges/vertices and make them always visible
    push_dir = (0, 0, 0)
    for ocp_face in faces:
        normal = Face(ocp_face).normal_at(v)
        push_dir = (push_dir[0] + normal.X, push_dir[1] + normal.Y, push_dir[2] + normal.Z)
    if push_dir != (0, 0, 0):
        # Normalize the push direction by the number of faces and a constant factor
        # NOTE: Don't overdo it, or metrics will be wrong
        n = len(faces) / 1e-3
        push_dir = (push_dir[0] / n, push_dir[1] / n, push_dir[2] / n)
        # Push the vertex by the normal
        v = (v[0] + push_dir[0], v[1] + push_dir[1], v[2] + push_dir[2])
    return v


def _tessellate_edge(
        mgr: GLTFMgr,
        ocp_edge: TopoDS_Edge,
        faces: List[TopoDS_Face],
        angular_deflection: float = 0.1,
        curvature_deflection: float = 0.1,
):
    # Use a curve discretizer to get the vertices
    curve = BRepAdaptor_Curve(ocp_edge)
    discretizer = GCPnts_TangentialDeflection(curve, angular_deflection, curvature_deflection)
    assert discretizer.NbPoints() > 1, "Edge is too small??"

    # add vertices
    vertices = [
        _push_point((v.X(), v.Y(), v.Z()), faces)
        for v in (
            discretizer.Value(i)  # .Transformed(transformation)
            for i in range(1, discretizer.NbPoints() + 1)
        )
    ]
    mgr.add_edge(vertices)


def _tessellate_vertex(mgr: GLTFMgr, ocp_vertex: TopoDS_Vertex, faces: List[TopoDS_Face]):
    c = Vertex(ocp_vertex).center()
    mgr.add_vertex(_push_point((c.X, c.Y, c.Z), faces))


def _hashcode(obj: Union[bytes, TopoDS_Shape], **extras) -> str:
    """Utility to compute the hash code of a shape recursively without the need to tessellate it"""
    # NOTE: obj.HashCode(MAX_HASH_CODE) is not stable across different runs of the same program
    # This is best-effort and not guaranteed to be unique
    hasher = hashlib.md5(usedforsecurity=False)
    for k, v in extras.items():
        hasher.update(str(k).encode())
        hasher.update(str(v).encode())
    if isinstance(obj, bytes):
        hasher.update(obj)
    elif isinstance(obj, TopLoc_Location):
        sub_data = io.BytesIO()
        obj.DumpJson(sub_data)
        hasher.update(sub_data.getvalue())
    elif isinstance(obj, TopoDS_Shape):
        map_of_shapes = TopTools_IndexedMapOfShape()
        TopExp.MapShapes_s(obj, map_of_shapes)
        for i in range(1, map_of_shapes.Extent() + 1):
            sub_shape = map_of_shapes.FindKey(i)
            sub_data = io.BytesIO()
            TopoDS_Shape.DumpJson(sub_shape, sub_data)
            val = sub_data.getvalue()
            val = re.sub(b'"this": "[^"]*"', b'', val)  # Remove memory address
            hasher.update(val)
    else:
        raise ValueError(f'Cannot hash object of type {type(obj)}')
    return hasher.hexdigest()



import os
from typing import Union, Dict

from build123d import *

ASSETS_DIR = os.getenv('ASSETS_DIR', os.path.join(os.path.dirname(__file__), '..', 'assets'))


def build_logo(text: bool = True) -> Dict[str, Union[Part, Location, str]]:
    """Builds the CAD part of the logo"""
    with BuildPart(Plane.XY.offset(50)) as logo_obj:
        Box(22, 40, 30)
        fillet(edges().filter_by(Axis.Y).group_by(Axis.Z)[-1], 10)
        offset(solid(), 2, openings=faces().group_by(Axis.Z)[0] + faces().filter_by(Plane.XZ))
        if text:
            text_at_plane = Plane.YZ
            text_at_plane.origin = faces().group_by(Axis.X)[-1].face().center()
            with BuildSketch(text_at_plane.location):
                Text('Yet Another\nCAD Viewer', 7, font_path='/usr/share/fonts/TTF/OpenSans-Regular.ttf')
            extrude(amount=1)

    logo_img_location = logo_obj.faces().group_by(Axis.X)[0].face().center_location  # Avoid overlapping:
    logo_img_location.position = Vector(logo_img_location.position.X - 4e-2, logo_img_location.position.Y,
                                        logo_img_location.position.Z)

    logo_img_path = os.path.join(ASSETS_DIR, 'img.jpg')
    img_bytes, img_name = prepare_image(logo_img_path, logo_img_location, height=18)

    fox_glb_bytes = open(os.path.join(ASSETS_DIR, 'fox.glb'), 'rb').read()

    return {'fox': fox_glb_bytes, 'logo': logo_obj, 'location': logo_img_location, img_name: img_bytes}


if __name__ == "__main__":
    import logging

    logging.basicConfig(level=logging.DEBUG)

    testing_server = os.getenv('TESTING_SERVER') is not None

    if not testing_server:
        # Start an offline server to export the CAD part of the logo in a way compatible with the frontend
        # If this is not set, the server will auto-start on import and show_* calls will provide live updates
        os.environ['YACV_DISABLE_SERVER'] = 'True'

    from yacv_server import export_all, remove, prepare_image, show

    # Build the CAD part of the logo
    logo = build_logo()

    # Add the CAD part of the logo to the server
    show(*[obj for obj in logo.values()], names=[name for name in logo.keys()])

    if testing_server:
        remove('location')  # Test removing a part
    else:
        # Save the complete logo to multiple GLB files
        export_all(os.path.join(ASSETS_DIR, 'logo_build'))
        print('Logo saved!')



import threading
import queue
import threading
from typing import List, TypeVar, \
    Generic, Generator

from yacv_server.mylogger import logger

T = TypeVar('T')


class BufferedPubSub(Generic[T]):
    """A simple implementation of publish-subscribe pattern using threading and buffering all previous events"""

    _buffer: List[T]
    _buffer_lock: threading.Lock
    _subscribers: List[queue.Queue[T]]
    _subscribers_lock: threading.Lock
    max_buffer_size: int

    def __init__(self, max_buffer_size: int = 100):
        self._buffer = []
        self._buffer_lock = threading.Lock()
        self._subscribers = []
        self._subscribers_lock = threading.Lock()
        self.max_buffer_size = max_buffer_size

    def publish(self, event: T):
        """Publishes an event without blocking (synchronous API does not require locking)"""
        with self._buffer_lock:
            self._buffer.append(event)
            if len(self._buffer) > self.max_buffer_size:
                self._buffer.pop(0)
            for q in self._subscribers:
                q.put(event)

    def _subscribe(self, include_buffered: bool = True, include_future: bool = True) -> queue.Queue[T]:
        """Subscribes to events"""
        q = queue.Queue()
        with self._subscribers_lock:
            self._subscribers.append(q)
        logger.debug(f"Subscribed to %s (%d subscribers)", self, len(self._subscribers))
        if include_buffered:
            with self._buffer_lock:
                for event in self._buffer:
                    q.put(event)
            if not include_future:
                q.put(None)
        return q

    def _unsubscribe(self, q: queue.Queue[T]):
        """Unsubscribes from events"""
        with self._subscribers_lock:
            self._subscribers.remove(q)
        logger.debug(f"Unsubscribed from %s (%d subscribers)", self, len(self._subscribers))

    def subscribe(self, include_buffered: bool = True, include_future: bool = True) -> Generator[T, None, None]:
        """Subscribes to events as an generator that yields events and automatically unsubscribes"""
        q = self._subscribe(include_buffered, include_future)
        try:
            while True:
                v = q.get()
                # include_future is incompatible with None values as they are used to signal the end of the stream
                if v is None and not include_future:
                    break
                yield v
        finally:  # When aclose() is called
            self._unsubscribe(q)

    def buffer(self) -> List[T]:
        """Returns a shallow copy of the list of buffered events"""
        with self._buffer_lock:
            return self._buffer[:]

    def delete(self, event: T):
        """Deletes an event from the buffer"""
        with self._buffer_lock:
            self._buffer.remove(event)

    def clear(self):
        """Clears the buffer"""
        with self._buffer_lock:
            self._buffer.clear()


from os import system

if __name__ == '__main__':
    # Just a reminder that a hot-reloading server can be started with the following command:
    # Need to disable auto-start to avoid conflicts with the hot-reloading server
    system('YACV_DISABLE_SERVER=true aiohttp-devtools runserver __init__.py --port 32323 --app-factory _get_app')



import os

from yacv_server.cad import image_to_gltf
from yacv_server.yacv import YACV

yacv = YACV()
"""The server instance. This is the main entry point to serve CAD objects and other data to the frontend."""

if 'YACV_DISABLE_SERVER' not in os.environ:
    # Start a new server ASAP to let the polling client connect while still building CAD objects
    # This is a bit of a hack, but it is seamless to the user. This behavior can be disabled by setting
    # the environment variable YACV_DISABLE_SERVER to a non-empty value
    yacv.start()

# Expose some nice aliases using the default server instance
show = yacv.show
show_all = yacv.show_cad_all
prepare_image = image_to_gltf
export_all = yacv.export_all
remove = yacv.remove
clear = yacv.clear



import io
import os
import threading
import urllib.parse
from http import HTTPStatus
from http.server import SimpleHTTPRequestHandler

from iterators import TimeoutIterator

from yacv_server.mylogger import logger

# Find the frontend folder (optional, but recommended)
FILE_DIR = os.path.dirname(__file__)
FRONTEND_BASE_PATH = os.getenv('FRONTEND_BASE_PATH', os.path.join(FILE_DIR, 'frontend'))
if not os.path.exists(FRONTEND_BASE_PATH):
    if os.path.exists(os.path.join(FILE_DIR, '..', 'dist')):  # Fallback to dev build
        FRONTEND_BASE_PATH = os.path.join(FILE_DIR, '..', 'dist')
    else:
        logger.warning('Frontend not found at %s', FRONTEND_BASE_PATH)
        FRONTEND_BASE_PATH = None

# Define the API paths (also available at the root path for simplicity)
UPDATES_API_PATH = '/api/updates'
OBJECTS_API_PATH = '/api/object'  # /{name}


class HTTPHandler(SimpleHTTPRequestHandler):
    yacv: 'yacv.YACV'
    frontend_lock: threading.Lock  # To avoid exiting too early while frontend makes requests
    at_least_one_client: threading.Event

    def __init__(self, *args, yacv: 'yacv.YACV', **kwargs):
        self.yacv = yacv
        self.frontend_lock = threading.Lock()
        self.at_least_one_client = threading.Event()
        super().__init__(*args, **kwargs, directory=FRONTEND_BASE_PATH)

    def log_message(self, fmt, *args):
        logger.debug(fmt, *args)

    def end_headers(self):
        # Add CORS headers to the response
        self.send_header('Access-Control-Allow-Origin', '*')
        self.send_header('Access-Control-Allow-Methods', 'GET, POST, OPTIONS')
        super().end_headers()

    def translate_path(self, path: str) -> str:
        """Translate a path to the local filesystem, adds some basic security checks"""
        path = super().translate_path(path)
        path = os.path.realpath(path)  # Avoid symlink hacks
        if self.directory:  # Ensure proper subdirectory
            base = os.path.abspath(self.directory)
            if not os.path.abspath(path).startswith(base):
                self.send_error(HTTPStatus.FORBIDDEN, "Path is not in the frontend directory")
                return ''
        return path

    def send_head(self):
        path_parts = self.path.split('?', 1)
        if len(path_parts) == 1:
            path_parts.append('')
        [path, query_str] = path_parts
        query = urllib.parse.parse_qs(query_str)
        if path == UPDATES_API_PATH or path == '/' and query.get('api_updates') is not None:
            return self._api_updates()
        elif path.startswith(OBJECTS_API_PATH) or path == '/' and query.get('api_object') is not None:
            if path.startswith(OBJECTS_API_PATH):
                obj_name = self.path[len(OBJECTS_API_PATH) + 1:]
            else:
                obj_name = query.get('api_object').pop()
            return self._api_object(obj_name)
        elif path.endswith('/'):  # Frontend index.html
            self.path += 'index.html'
            return super().send_head()
        else:  # Normal frontend file
            return super().send_head()

    def _api_updates(self):
        """Handles a publish-only websocket connection that send show_object events along with their hashes and URLs"""
        self.send_response(HTTPStatus.OK)
        self.send_header("Content-Type", "text/event-stream")
        self.send_header("Cache-Control", "no-cache")
        # Chunked transfer encoding!
        self.send_header("Transfer-Encoding", "chunked")
        self.end_headers()
        self.at_least_one_client.set()
        logger.debug('Updates client connected')

        def write_chunk(_chunk_data: str):
            self.wfile.write(hex(len(_chunk_data))[2:].encode('utf-8'))
            self.wfile.write(b'\r\n')
            self.wfile.write(_chunk_data.encode('utf-8'))
            self.wfile.write(b'\r\n')
            self.wfile.flush()

        write_chunk('retry: 100\n\n')

        # Send buffered events first, while keeping a lock
        with self.frontend_lock:
            for data in self.yacv.show_events.buffer():
                logger.debug('Sending info about %s: %s', data.name, data)
                # noinspection PyUnresolvedReferences
                to_send = data.to_json()
                write_chunk(f'data: {to_send}\n\n')

        # Send future events over the same connection
        # Also send keep-alive to know if the client is still connected
        subscription = self.yacv.show_events.subscribe(include_buffered=False)
        it = TimeoutIterator(subscription, sentinel=None, reset_on_next=True, timeout=5.0)  # Keep-alive interval
        try:
            for data in it:
                if data is None:
                    write_chunk(':keep-alive\n\n')
                else:
                    logger.debug('Sending info about %s: %s', data.name, data)
                    # noinspection PyUnresolvedReferences
                    to_send = data.to_json()
                    write_chunk(f'data: {to_send}\n\n')
        except BrokenPipeError:  # Client disconnected normally
            pass
        finally:
            logger.debug('Updates client disconnected')
            try:
                it.interrupt()
                next(it)  # Make sure the iterator is interrupted before trying to close the subscription
                subscription.close()
            except BaseException as e:
                logger.debug('Ignoring error while closing subscription: %s', e)

    def _api_object(self, obj_name: str):
        """Returns the object file with the matching name, building it if necessary."""
        with self.frontend_lock:
            # Export the object (or fail if not found)
            exported_glb = self.yacv.export(obj_name)
            if exported_glb is None:
                self.send_error(HTTPStatus.NOT_FOUND, f'Object {obj_name} not found')
                return io.BytesIO()

            # Wrap the GLB in a response and return it
            self.send_response(HTTPStatus.OK)
            self.send_header('Content-Type', 'model/gltf-binary')
            self.send_header('Content-Length', str(len(exported_glb)))
            self.send_header('Content-Disposition', f'attachment; filename="{obj_name}.glb"')
            self.end_headers()
            self.wfile.write(exported_glb)



import atexit
import copy
import inspect
import os
import signal
import sys
import threading
import time
from dataclasses import dataclass
from http.server import ThreadingHTTPServer
from importlib.metadata import version
from threading import Thread
from typing import Optional, Dict, Union, Callable, List

from OCP.TopLoc import TopLoc_Location
from OCP.TopoDS import TopoDS_Shape
# noinspection PyProtectedMember
from build123d import Shape, Axis, Location, Vector
from dataclasses_json import dataclass_json

from yacv_server.cad import get_shape, grab_all_cad, CADCoreLike, CADLike
from yacv_server.myhttp import HTTPHandler
from yacv_server.mylogger import logger
from yacv_server.pubsub import BufferedPubSub
from yacv_server.tessellate import _hashcode, tessellate


@dataclass_json
@dataclass
class UpdatesApiData:
    """Data sent to the client through the updates API"""
    name: str
    """Name of the object. Should be unique unless you want to overwrite the previous object"""
    hash: str
    """Hash of the object, to detect changes without rebuilding the object"""
    is_remove: bool
    """Whether to remove the object from the scene"""


YACVSupported = Union[bytes, CADCoreLike]


class UpdatesApiFullData(UpdatesApiData):
    obj: YACVSupported
    """The OCCT object, if any (not serialized)"""
    kwargs: Optional[Dict[str, any]]
    """The show_object options, if any (not serialized)"""

    def __init__(self, obj: YACVSupported, name: str, _hash: str, is_remove: bool = False,
                 kwargs: Optional[Dict[str, any]] = None):
        self.name = name
        self.hash = _hash
        self.is_remove = is_remove
        self.obj = obj
        self.kwargs = kwargs

    def to_json(self) -> str:
        # noinspection PyUnresolvedReferences
        return super().to_json()


class YACV:
    server_thread: Optional[Thread]
    server: Optional[ThreadingHTTPServer]
    startup_complete: threading.Event
    show_events: BufferedPubSub[UpdatesApiFullData]
    build_events: Dict[str, BufferedPubSub[bytes]]
    object_events_lock: threading.Lock

    def __init__(self):
        self.server_thread = None
        self.server = None
        self.startup_complete = threading.Event()
        self.at_least_one_client = threading.Event()
        self.show_events = BufferedPubSub()
        self.build_events = {}
        self.object_events_lock = threading.Lock()
        self.frontend_lock = threading.Lock()
        logger.info('Using yacv-server v%s', version('yacv-server'))

    def start(self):
        """Starts the web server in the background"""
        assert self.server_thread is None, "Server currently running, cannot start another one"
        assert self.startup_complete.is_set() is False, "Server already started"
        # Start the server in a separate daemon thread
        self.server_thread = Thread(target=self._run_server, name='yacv_server', daemon=True)
        signal.signal(signal.SIGINT | signal.SIGTERM, self.stop)
        atexit.register(self.stop)
        self.server_thread.start()
        logger.info('Server started (requested)...')
        # Wait for the server to be ready before returning
        while not self.startup_complete.wait():
            time.sleep(0.01)
        logger.info('Server started (received)...')

    # noinspection PyUnusedLocal
    def stop(self, *args):
        """Stops the web server"""
        if self.server_thread is None:
            logger.error('Cannot stop server because it is not running')
            return

        graceful_secs_connect = float(os.getenv('YACV_GRACEFUL_SECS_CONNECT', 12.0))
        graceful_secs_request = float(os.getenv('YACV_GRACEFUL_SECS_REQUEST', 5.0))
        # Make sure we can hold the lock for more than 100ms (to avoid exiting too early)
        logger.info('Stopping server (waiting for at least one frontend request first, cancel with CTRL+C)...')
        start = time.time()
        try:
            while not self.at_least_one_client.wait(
                    graceful_secs_connect / 10) and time.time() - start < graceful_secs_connect:
                time.sleep(0.01)
        except KeyboardInterrupt:
            pass

        logger.info('Stopping server (waiting for no more frontend requests)...')
        start = time.time()
        try:
            while time.time() - start < graceful_secs_request:
                if self.frontend_lock.locked():
                    start = time.time()
                time.sleep(0.01)
        except KeyboardInterrupt:
            pass

        # Stop the server in the background
        self.server.shutdown()
        logger.info('Stopping server (sent)...')

        # Wait for the server to stop gracefully
        self.server_thread.join(timeout=30)
        self.server_thread = None
        logger.info('Stopping server (confirmed)...')
        if len(args) >= 1 and args[0] in (signal.SIGINT, signal.SIGTERM):
            sys.exit(0)  # Exit with success

    def _run_server(self):
        """Runs the web server"""
        logger.info('Starting server...')
        self.server = ThreadingHTTPServer(
            (os.getenv('YACV_HOST', 'localhost'), int(os.getenv('YACV_PORT', 32323))),
            lambda a, b, c: HTTPHandler(a, b, c, yacv=self))
        # noinspection HttpUrlsUsage
        logger.info(f'Serving at http://{self.server.server_name}:{self.server.server_port}')
        self.startup_complete.set()
        self.server.serve_forever()

    def show(self, *objs: List[YACVSupported], names: Optional[Union[str, List[str]]] = None, **kwargs):
        # Prepare the arguments
        start = time.time()
        names = names or [_find_var_name(obj) for obj in objs]
        if isinstance(names, str):
            names = [names]
        assert len(names) == len(objs), 'Number of names must match the number of objects'

        # Handle auto clearing of previous objects
        if kwargs.get('auto_clear', True):
            self.clear(except_names=names)

        # Remove a previous object event with the same name
        for old_event in self.show_events.buffer():
            if old_event.name in names:
                self.show_events.delete(old_event)
                if old_event.name in self.build_events:
                    del self.build_events[old_event.name]

        # Publish the show event
        for obj, name in zip(objs, names):
            if not isinstance(obj, bytes):
                obj = _preprocess_cad(obj, **kwargs)
            _hash = _hashcode(obj, **kwargs)
            event = UpdatesApiFullData(name=name, _hash=_hash, obj=obj, kwargs=kwargs or {})
            self.show_events.publish(event)

        logger.info('show %s took %.3f seconds', names, time.time() - start)

    def show_cad_all(self, **kwargs):
        """Publishes all CAD objects in the current scope to the server"""
        all_cad = grab_all_cad()
        self.show(*[cad for _, cad in all_cad], names=[name for name, _ in all_cad], **kwargs)

    def remove(self, name: str):
        """Removes a previously-shown object from the scene"""
        show_events = self._show_events(name)
        if len(show_events) > 0:
            # Ensure only the new remove event remains for this name
            for old_show_event in show_events:
                self.show_events.delete(old_show_event)

            # Delete any cached object builds
            with self.object_events_lock:
                if name in self.build_events:
                    del self.build_events[name]

            # Publish the remove event
            show_event = copy.copy(show_events[-1])
            show_event.is_remove = True
            self.show_events.publish(show_event)

    def clear(self, except_names: List[str] = None):
        """Clears all previously-shown objects from the scene"""
        if except_names is None:
            except_names = []
        for event in self.show_events.buffer():
            if event.name not in except_names:
                self.remove(event.name)

    def shown_object_names(self, apply_removes: bool = True) -> List[str]:
        """Returns the names of all objects that have been shown"""
        res = set()
        for obj in self.show_events.buffer():
            if not obj.is_remove or not apply_removes:
                res.add(obj.name)
            else:
                res.discard(obj.name)
        return list(res)

    def _show_events(self, name: str, apply_removes: bool = True) -> List[UpdatesApiFullData]:
        """Returns the show events with the given name"""
        res = []
        for event in self.show_events.buffer():
            if event.name == name:
                if not event.is_remove or not apply_removes:
                    res.append(event)
                else:
                    # Also remove the previous events
                    for old_event in res:
                        if old_event.name == event.name:
                            res.remove(old_event)
        return res

    def export(self, name: str) -> Optional[bytes]:
        """Export the given previously-shown object to a single GLB file, building it if necessary."""
        start = time.time()

        # Check that the object to build exists and grab it if it does
        events = self._show_events(name)
        if len(events) == 0:
            logger.warning('Object %s not found', name)
            return None
        event = events[-1]

        # Use the lock to ensure that we don't build the object twice
        with self.object_events_lock:
            # If there are no object events for this name, we need to build the object
            if name not in self.build_events:
                logger.debug('Building object %s with hash %s', name, event.hash)

                # Prepare the pubsub for the object
                publish_to = BufferedPubSub[bytes]()
                self.build_events[name] = publish_to

                # Build and publish the object (once)
                if isinstance(event.obj, bytes):  # Already a GLTF
                    publish_to.publish(event.obj)
                else:  # CAD object to tessellate and convert to GLTF
                    gltf = tessellate(event.obj, tolerance=event.kwargs.get('tolerance', 0.1),
                                      angular_tolerance=event.kwargs.get('angular_tolerance', 0.1),
                                      faces=event.kwargs.get('faces', True),
                                      edges=event.kwargs.get('edges', True),
                                      vertices=event.kwargs.get('vertices', True))
                    glb_list_of_bytes = gltf.save_to_bytes()
                    publish_to.publish(b''.join(glb_list_of_bytes))
                    logger.info('export(%s) took %.3f seconds, %d parts', name, time.time() - start,
                                len(gltf.meshes[0].primitives))

            # In either case return the elements of a subscription to the async generator
            subscription = self.build_events[name].subscribe()
            try:
                return next(subscription)
            finally:
                subscription.close()

    def export_all(self, folder: str,
                   export_filter: Callable[[str, Optional[CADCoreLike]], bool] = lambda name, obj: True):
        """Export all previously-shown objects to GLB files in the given folder"""
        os.makedirs(folder, exist_ok=True)
        for name in self.shown_object_names():
            if export_filter(name, self._show_events(name)[-1].obj):
                with open(os.path.join(folder, f'{name}.glb'), 'wb') as f:
                    f.write(self.export(name))


# noinspection PyUnusedLocal
def _preprocess_cad(obj: CADLike, **kwargs) -> CADCoreLike:
    # Get the shape of a CAD-like object
    obj = get_shape(obj)

    # Convert Z-up (OCCT convention) to Y-up (GLTF convention)
    if isinstance(obj, TopoDS_Shape):
        obj = Shape(obj).rotate(Axis.X, -90).wrapped
    elif isinstance(obj, TopLoc_Location):
        tmp_location = Location(obj)
        tmp_location.position = Vector(tmp_location.position.X, tmp_location.position.Z,
                                       -tmp_location.position.Y)
        tmp_location.orientation = Vector(tmp_location.orientation.X - 90, tmp_location.orientation.Y,
                                          tmp_location.orientation.Z)
        obj = tmp_location.wrapped

    return obj


_find_var_name_count = 0


def _find_var_name(obj: any, avoid_levels: int = 2) -> str:
    """A hacky way to get a stable name for an object that may change over time"""
    global _find_var_name_count
    for frame in inspect.stack()[avoid_levels:]:
        for key, value in frame.frame.f_locals.items():
            if value is obj:
                return key
    _find_var_name_count += 1
    return 'unknown_var_' + str(_find_var_name_count)



"""
Utilities to work with CAD objects
"""
import hashlib
from typing import Optional, Union, List, Tuple

from OCP.TopLoc import TopLoc_Location
from OCP.TopoDS import TopoDS_Shape

from yacv_server.gltf import GLTFMgr

CADCoreLike = Union[TopoDS_Shape, TopLoc_Location]  # Faces, Edges, Vertices and Locations for now
CADLike = Union[CADCoreLike, any]  # build123d and cadquery types


def get_shape(obj: CADLike, error: bool = True) -> Optional[CADCoreLike]:
    """ Get the shape of a CAD-like object """

    # Try to grab a shape if a different type of object was passed
    if isinstance(obj, TopoDS_Shape) or isinstance(obj, TopLoc_Location):
        return obj

    # Return locations (drawn as axes)
    if 'wrapped' in dir(obj) and isinstance(obj.wrapped, TopLoc_Location):
        return obj.wrapped

    # Build123D
    if 'part' in dir(obj):
        obj = obj.part
    if 'sketch' in dir(obj):
        obj = obj.sketch
    if 'line' in dir(obj):
        obj = obj.line

    # Build123D & CadQuery
    while 'wrapped' in dir(obj) and not isinstance(obj, TopoDS_Shape) and not isinstance(obj, TopLoc_Location):
        obj = obj.wrapped

    # Return shapes
    if isinstance(obj, TopoDS_Shape):
        return obj

    if error:
        raise ValueError(f'Cannot show object of type {type(obj)} (submit issue?)')
    else:
        return None


def grab_all_cad() -> List[Tuple[str, CADCoreLike]]:
    """ Grab all shapes by inspecting the stack """
    import inspect
    stack = inspect.stack()
    shapes = []
    for frame in stack:
        for key, value in frame.frame.f_locals.items():
            shape = get_shape(value, error=False)
            if shape:
                shapes.append((key, shape))
    return shapes


def image_to_gltf(source: str | bytes, center: any, width: Optional[float] = None, height: Optional[float] = None,
                  name: Optional[str] = None, save_mime: str = 'image/jpeg') -> Tuple[bytes, str]:
    """Convert an image to a GLTF CAD object."""
    from PIL import Image
    import io
    import os
    from build123d import Plane
    from build123d import Location
    from build123d import Vector

    # Handle arguments
    if name is None:
        if isinstance(source, str):
            name = os.path.basename(source)
        else:
            hasher = hashlib.md5()
            hasher.update(source)
            name = 'image_' + hasher.hexdigest()
    format: str
    if save_mime == 'image/jpeg':
        format = 'JPEG'
    elif save_mime == 'image/png':
        format = 'PNG'
    else:
        raise ValueError(f'Unsupported save MIME type (for GLTF files): {save_mime}')

    # Get the plane of the image
    center_loc = get_shape(center)
    if not isinstance(center_loc, TopLoc_Location):
        raise ValueError('Center location not valid')
    plane = Plane(Location(center_loc))
    # Convert coordinates system
    plane.origin = Vector(plane.origin.X, plane.origin.Z, -plane.origin.Y)
    plane.z_dir = -plane.y_dir
    plane.y_dir = plane.z_dir

    def vert(v: Vector) -> Tuple[float, float, float]:
        return v.X, v.Y, v.Z

    # Load the image to a byte buffer
    img = Image.open(source)
    img_buf = io.BytesIO()
    img.save(img_buf, format=format)
    img_buf = img_buf.getvalue()

    # Build the gltf
    mgr = GLTFMgr(image=(img_buf, save_mime))
    if width is None and height is None:
        raise ValueError('At least one of width or height must be specified')  # Fallback to pixels == mm?
    elif width is None:
        width = img.width / img.height * height
    elif height is None:
        height = height or img.height / img.width * width  # Apply default aspect ratio if unspecified
    mgr.add_face([
        vert(plane.origin - plane.x_dir * width / 2 - plane.y_dir * height / 2),
        vert(plane.origin + plane.x_dir * width / 2 - plane.y_dir * height / 2),
        vert(plane.origin + plane.x_dir * width / 2 + plane.y_dir * height / 2),
        vert(plane.origin - plane.x_dir * width / 2 + plane.y_dir * height / 2),
    ], [
        (0, 2, 1),
        (0, 3, 2),
    ], [
        (0, 0),
        (1, 0),
        (1, 1),
        (0, 1),
    ])

    # Return the GLTF binary blob and the suggested name of the image
    return b''.join(mgr.gltf.save_to_bytes()), name



import importlib.metadata

import numpy as np
from build123d import Location, Plane, Vector
from pygltflib import *

_checkerboard_image_bytes = base64.decodebytes(
    b'iVBORw0KGgoAAAANSUhEUgAAAAIAAAACCAIAAAD91JpzAAAAF0lEQVQI12N49OjR////Gf'
    b'/////48WMATwULS8tcyj8AAAAASUVORK5CYII=')


class GLTFMgr:
    """A utility class to build our GLTF2 objects easily and incrementally"""

    def __init__(self, image: Tuple[bytes, str] = (_checkerboard_image_bytes, 'image/png')):
        self.gltf = GLTF2(
            asset=Asset(generator=f"yacv_server@{importlib.metadata.version('yacv_server')}"),
            scene=0,
            scenes=[Scene(nodes=[0])],
            nodes=[Node(mesh=0)],
            meshes=[Mesh(primitives=[])],
            accessors=[],
            bufferViews=[BufferView(buffer=0, byteLength=len(image[0]), byteOffset=0)],
            buffers=[Buffer(byteLength=len(image[0]))],
            samplers=[Sampler(magFilter=NEAREST)],
            textures=[Texture(source=0, sampler=0)],
            images=[Image(bufferView=0, mimeType=image[1])],
        )
        # TODO: Reduce the number of draw calls by merging all faces into a single primitive, and using
        #  color attributes + extension? to differentiate them (same for edges and vertices)
        self.gltf.set_binary_blob(image[0])

    def add_face(self, vertices_raw: List[Tuple[float, float, float]], indices_raw: List[Tuple[int, int, int]],
                 tex_coord_raw: List[Tuple[float, float]]):
        """Add a face to the GLTF as a new primitive of the unique mesh"""
        vertices = np.array([[v[0], v[1], v[2]] for v in vertices_raw], dtype=np.float32)
        indices = np.array([[i[0], i[1], i[2]] for i in indices_raw], dtype=np.uint32)
        tex_coord = np.array([[t[0], t[1]] for t in tex_coord_raw], dtype=np.float32)
        self._add_any(vertices, indices, tex_coord, mode=TRIANGLES, material="face")

    def add_edge(self, vertices_raw: List[Tuple[float, float, float]], mat: str = None):
        """Add an edge to the GLTF as a new primitive of the unique mesh"""
        vertices = np.array([[v[0], v[1], v[2]] for v in vertices_raw], dtype=np.float32)
        indices = np.array(list(map(lambda i: [i, i + 1], range(len(vertices) - 1))), dtype=np.uint32)
        tex_coord = np.array([])
        self._add_any(vertices, indices, tex_coord, mode=LINE_STRIP, material=mat or "edge")

    def add_vertex(self, vertex: Tuple[float, float, float]):
        """Add a vertex to the GLTF as a new primitive of the unique mesh"""
        vertices = np.array([[vertex[0], vertex[1], vertex[2]]])
        indices = np.array([[0]], dtype=np.uint32)
        tex_coord = np.array([], dtype=np.float32)
        self._add_any(vertices, indices, tex_coord, mode=POINTS, material="vertex")

    def add_location(self, loc: Location):
        """Add a location to the GLTF as a new primitive of the unique mesh"""
        pl = Plane(loc)

        def vert(v: Vector) -> Tuple[float, float, float]:
            return v.X, v.Y, v.Z

        # Add 1 origin vertex and 3 edges with custom colors to identify the X, Y and Z axis
        self.add_vertex(vert(pl.origin))
        self.add_edge([vert(pl.origin), vert(pl.origin + pl.x_dir)], mat="locX")
        self.add_edge([vert(pl.origin), vert(pl.origin + pl.y_dir)], mat="locY")
        self.add_edge([vert(pl.origin), vert(pl.origin + pl.z_dir)], mat="locZ")

    def add_material(self, kind: str) -> int:
        """It is important to use a different material for each primitive to be able to change them at runtime"""
        new_material: Material
        if kind == "face":
            new_material = Material(name="face", alphaCutoff=None, pbrMetallicRoughness=PbrMetallicRoughness(
                baseColorTexture=TextureInfo(index=0), baseColorFactor=[1, 1, 0.5, 1]), doubleSided=True)
        elif kind == "edge":
            new_material = Material(name="edge", alphaCutoff=None, pbrMetallicRoughness=PbrMetallicRoughness(
                baseColorFactor=[0, 0, 0.5, 1]))
        elif kind == "vertex":
            new_material = Material(name="vertex", alphaCutoff=None, pbrMetallicRoughness=PbrMetallicRoughness(
                baseColorFactor=[0, 0.3, 0.3, 1]))
        elif kind == "locX":
            new_material = Material(name="locX", alphaCutoff=None, pbrMetallicRoughness=PbrMetallicRoughness(
                baseColorFactor=[0.97, 0.24, 0.24, 1]))
        elif kind == "locY":
            new_material = Material(name="locY", alphaCutoff=None, pbrMetallicRoughness=PbrMetallicRoughness(
                baseColorFactor=[0.42, 0.8, 0.15, 1]))
        elif kind == "locZ":
            new_material = Material(name="locZ", alphaCutoff=None, pbrMetallicRoughness=PbrMetallicRoughness(
                baseColorFactor=[0.09, 0.55, 0.94, 1]))
        else:
            raise ValueError(f"Unknown material kind {kind}")
        self.gltf.materials.append(new_material)
        return len(self.gltf.materials) - 1

    def _add_any(self, vertices: np.ndarray, indices: np.ndarray, tex_coord: np.ndarray, mode: int = TRIANGLES,
                 material: str = "face"):
        assert vertices.ndim == 2
        assert vertices.shape[1] == 3
        vertices = vertices.astype(np.float32)
        vertices_blob = vertices.tobytes()

        assert indices.ndim == 2
        assert indices.shape[1] == 3 and mode == TRIANGLES or indices.shape[1] == 2 and mode == LINE_STRIP or \
               indices.shape[1] == 1 and mode == POINTS
        indices = indices.astype(np.uint32)
        indices_blob = indices.flatten().tobytes()

        # Check that all vertices are referenced by the indices
        # This can happen on broken faces like on some fonts
        # assert indices.max() == len(vertices) - 1, f"{indices.max()} != {len(vertices) - 1}"
        # assert indices.min() == 0, f"min({indices}) != 0"
        # assert np.unique(indices.flatten()).size == len(vertices)

        assert len(tex_coord) == 0 or tex_coord.ndim == 2
        assert len(tex_coord) == 0 or tex_coord.shape[1] == 2
        tex_coord = tex_coord.astype(np.float32)
        tex_coord_blob = tex_coord.tobytes()

        accessor_base = len(self.gltf.accessors)
        self.gltf.meshes[0].primitives.append(
            Primitive(
                attributes=Attributes(POSITION=accessor_base + 1, TEXCOORD_0=accessor_base + 2)
                if len(tex_coord) > 0 else Attributes(POSITION=accessor_base + 1),
                indices=accessor_base,
                mode=mode,
                material=self.add_material(material),
            )
        )

        buffer_view_base = len(self.gltf.bufferViews)
        self.gltf.accessors.extend([it for it in [
            Accessor(
                bufferView=buffer_view_base,
                componentType=UNSIGNED_INT,
                count=indices.size,
                type=SCALAR,
                max=[int(indices.max())],
                min=[int(indices.min())],
            ),
            Accessor(
                bufferView=buffer_view_base + 1,
                componentType=FLOAT,
                count=len(vertices),
                type=VEC3,
                max=vertices.max(axis=0).tolist(),
                min=vertices.min(axis=0).tolist(),
            ),
            Accessor(
                bufferView=buffer_view_base + 2,
                componentType=FLOAT,
                count=len(tex_coord),
                type=VEC2,
                max=tex_coord.max(axis=0).tolist(),
                min=tex_coord.min(axis=0).tolist(),
            ) if len(tex_coord) > 0 else None
        ] if it is not None])

        prev_binary_blob = self.gltf.binary_blob()
        byte_offset_base = len(prev_binary_blob)
        self.gltf.bufferViews.extend([bv for bv in [
            BufferView(
                buffer=0,
                byteOffset=byte_offset_base,
                byteLength=len(indices_blob),
                target=ELEMENT_ARRAY_BUFFER,
            ),
            BufferView(
                buffer=0,
                byteOffset=byte_offset_base + len(indices_blob),
                byteLength=len(vertices_blob),
                target=ARRAY_BUFFER,
            ),
            BufferView(
                buffer=0,
                byteOffset=byte_offset_base + len(indices_blob) + len(vertices_blob),
                byteLength=len(tex_coord_blob),
                target=ARRAY_BUFFER,
            )
        ] if bv.byteLength > 0])

        self.gltf.set_binary_blob(prev_binary_blob + indices_blob + vertices_blob + tex_coord_blob)
